/*
 * MSC (MMC/SD controller) peripheral support.
 *
 * Copyright (C) 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include <l4/sys/irq.h>
#include <l4/sys/rcv_endpoint.h>
#include <l4/util/util.h>

#include <systypes/thread.h>

#include <math.h>
#include <stdio.h>

#include "msc-common.h"
#include "msc-defs.h"



// Command frame:
// byte: start (1), direction (1), command (6)
// 4 bytes: argument
// byte: CRC (7), end (1)

// IO_RW_DIRECT argument:
// Argument MSB to LSB: R/W (1), function number (3), read after write flag (1),
//                      stuff (1), register address (17), stuff (1),
//                      write data or stuff (8)
// 0x88000c08: W, function = 0, read after write, register address = 6 (CCCR),
//             data = 8 (reset)

const uint32_t Io_rw_direct_reset = 0x88000c08;

// (IO_)SEND_OP_COND argument and default voltage range expected in R3, R4:
// Argument MSB to LSB: stuff (8), voltage range (16), reserved (8)
// 0x00ff8000: voltage range 2.7 - 3.6V

const uint32_t Ocr_default_voltage_range = 0x00ff8000;

// SEND_IF_COND argument and default voltage range expected in R7:
// Argument MSB to LSB: stuff (20), voltage supplied (4), check (8)
// 0x000001aa: voltage range 2.7 - 3.6V, check = 0b10101010

const uint32_t If_cond_default_voltage_range = 0x000001aa;



// Utilities.

static enum Command_data_control_bits
encode_bus_width(uint8_t width)
{
  switch (width)
  {
    case 4: return Cdc_bus_width_4bit;
    case 1: return Cdc_bus_width_1bit;
    default: return Cdc_bus_width_1bit;
  }
}



// Channel abstraction.

Msc_channel::Msc_channel(l4_addr_t msc_start, l4_addr_t addr, l4_cap_idx_t irq,
                         Cpm_chip *cpm, enum Clock_identifiers clock)
: _msc_start(msc_start), _irq(irq), _cpm(cpm), _clock(clock)
{
  _regs = new Hw::Mmio_register_block<32>(addr);
}

Msc_channel::~Msc_channel()
{
}

// Utility methods.
// NOTE: Also defined in the CPM abstraction, should be consolidated.

uint32_t
Msc_channel::get_field(uint32_t reg, uint32_t mask, uint8_t shift)
{
  return (_regs[reg] & (mask << shift)) >> shift;
}

void
Msc_channel::set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value)
{
  _regs[reg] = (_regs[reg] & (~(mask << shift))) | ((mask & value) << shift);
}

bool
Msc_channel::set_clock_frequency(uint64_t frequency)
{
  uint64_t division = _cpm->get_frequency(_clock) / frequency;
  double divider = ceil(log2(division));

  if ((divider < 0) || (divider > Clock_rate_field_mask))
    return false;

  set_field(Msc_clock_rate, Clock_rate_field_mask, Clock_rate_field_shift,
            (uint32_t) divider);

  return true;
}

void 
Msc_channel::reset()
{
  _regs[Msc_control] = _regs[Msc_control] | Control_reset;

  while (_regs[Msc_status] & Status_resetting);
}

void
Msc_channel::start_clock()
{
  set_field(Msc_control, Control_clock_control_field_mask,
            Control_clock_control_field_shift, Control_clock_control_start);

  while (!(_regs[Msc_status] & Status_clock_enabled));
}

void
Msc_channel::stop_clock()
{
  set_field(Msc_control, Control_clock_control_field_mask,
            Control_clock_control_field_shift, Control_clock_control_stop);

  while (_regs[Msc_status] & Status_clock_enabled);
}

void 
Msc_channel::ack_irq(uint32_t flags)
{
  // Clear the flags by setting them.

  _regs[Msc_interrupt_flag] = _regs[Msc_interrupt_flag] | flags;
}

void 
Msc_channel::unmask_irq(uint32_t flags)
{
  ack_irq(flags);

  if (_regs[Msc_interrupt_mask] & flags)
    _regs[Msc_interrupt_mask] = _regs[Msc_interrupt_mask] & ~flags;
}

bool
Msc_channel::have_dma_enable_in_command()
{
  return false;
}

bool
Msc_channel::have_dma_selection()
{
  return true;
}

bool
Msc_channel::command_will_write(uint8_t index)
{
  // NOTE: Probably incomplete coverage.

  switch (index)
  {
    case Command_write_dat_until_stop: return true;
    case Command_write_block: return true;
    case Command_write_multiple_block: return true;
    case Command_program_cid: return true;
    case Command_program_csd: return true;
    case Command_lock_unlock: return true;
    default: return false;
  }
}

bool
Msc_channel::app_command_will_write(uint8_t index)
{
  // NOTE: Probably incomplete coverage.

  (void) index;

  return false;
}

bool
Msc_channel::command_with_data(uint8_t index)
{
  // NOTE: Probably incomplete coverage.

  switch (index)
  {
    case Command_read_dat_until_stop: return true;
    case Command_read_single_block: return true;
    case Command_read_multiple_block: return true;
    case Command_write_dat_until_stop: return true;
    case Command_write_block: return true;
    case Command_write_multiple_block: return true;
    case Command_program_cid: return true;
    case Command_program_csd: return true;
    case Command_lock_unlock: return true;
    default: return false;
  }
}

bool
Msc_channel::app_command_with_data(uint8_t index)
{
  // NOTE: Probably incomplete coverage.

  switch (index)
  {
    case App_command_sd_status: return true;
    case App_command_send_num_wr_blocks: return true;
    case App_command_send_scr: return true;
    default: return false;
  }
}

bool
Msc_channel::command_uses_busy(uint8_t index)
{
  // NOTE: Probably incomplete coverage.

  switch (index)
  {
    case Command_select_deselect_card: return true;
    case Command_stop_transmission: return true;
    default: return false;
  }
}

bool
Msc_channel::app_command_uses_busy(uint8_t index)
{
  // NOTE: Probably incomplete coverage.

  (void) index;

  return false;
}

uint8_t
Msc_channel::get_response_format(uint8_t index)
{
  // NOTE: Probably incomplete coverage.

  switch (index)
  {
    // Common commands without response.

    case Command_go_idle_state: return 0;
    case Command_set_dsr: return 0;
    case Command_go_inactive_state: return 0;

    // Common commands with response.

    case Command_send_op_cond: return 3;
    case Command_all_send_cid: return 2;
    case Command_send_csd: return 2;
    case Command_send_cid: return 2;

    // SDIO only.

    case Command_io_send_op_cond: return 4;
    case Command_io_rw_direct: return 5;

    // SDMEM only.

    case Command_send_relative_addr: return 6;
    case Command_send_if_cond: return 7;

    // All other commands.

    default: return 1;
  }
}

uint8_t
Msc_channel::get_app_response_format(uint8_t index)
{
  // NOTE: Probably incomplete coverage.

  switch (index)
  {
    // SDMEM only.

    case App_command_sd_send_op_cond: return 3;

    // All other commands.

    default: return 1;
  }
}

// Read a response directly from the FIFO.

void
Msc_channel::read_response(uint16_t *buffer, uint8_t units)
{
  uint8_t unit = units;

  while (unit > 0)
  {
    uint32_t data = _regs[Msc_response_fifo];

    // Ignore the upper byte of the last unit in small transfers since it is the
    // lower byte from the previous unit not shifted out of the register.

    unit--;

    if ((unit == 0) && (units == 3))
      buffer[unit] = (data & 0xff) << 8;
    else
      buffer[unit] = data;
  }
}

// Send an application-specific command.

bool
Msc_channel::send_app_command(uint8_t index, uint32_t arg)
{
  if (!send_command(Command_app_cmd, _cards[_card].rca << 16))
    return false;

  return send_command(index, arg, get_app_response_format(index),
                      app_command_with_data(index),
                      app_command_will_write(index),
                      app_command_uses_busy(index));
}

// Send a common MMC/SD command.

bool
Msc_channel::send_command(uint8_t index, uint32_t arg)
{
  return send_command(index, arg, get_response_format(index),
                      command_with_data(index),
                      command_will_write(index),
                      command_uses_busy(index));
}

// Initiate a command having the given index and using the given argument,
// employing the specified response format and involving a data transfer if
// indicated.

bool
Msc_channel::send_command(uint8_t index, uint32_t arg, uint8_t response_format,
                          bool data, bool write, bool busy)
{
  stop_clock();

  // Enable DMA for data transfers for JZ4780, X1600 and others that have a
  // channel-level control selecting special or conventional DMA.

  if (have_dma_selection())
    _regs[Msc_dma_control] = (data ? Dma_select_common_dma | Dma_enable : Dma_disable);

  // Set the command index and argument.

  _regs[Msc_command_index] = index;
  _regs[Msc_command_argument] = arg;

  // Configure the response format and data bus width.

  set_field(Msc_command_data_control, Cdc_response_format_field_mask,
            Cdc_response_format_field_shift, response_format);

  // NOTE: May need to set the SD bus width.

  set_field(Msc_command_data_control, Cdc_bus_width_field_mask,
            Cdc_bus_width_field_shift,
            encode_bus_width(_cards[_card].bus_width));

  set_field(Msc_command_data_control, Cdc_recv_fifo_level_field_mask,
            Cdc_recv_fifo_level_field_shift, Cdc_fifo_level_16);

  set_field(Msc_command_data_control, Cdc_trans_fifo_level_field_mask,
            Cdc_trans_fifo_level_field_shift, Cdc_fifo_level_16);

  // Set and clear control bits appropriate to the command.

  _regs[Msc_command_data_control] = _regs[Msc_command_data_control] |
    (busy ? Cdc_expect_busy : Cdc_do_not_expect_busy) |
    (data ? Cdc_data_with_command : Cdc_no_data_with_command) |
    (write ? Cdc_write_operation : Cdc_read_operation);

  _regs[Msc_command_data_control] = _regs[Msc_command_data_control] &
    ~((busy ? Cdc_do_not_expect_busy : Cdc_expect_busy) |
      (data ? Cdc_no_data_with_command : Cdc_data_with_command) |
      (write ? Cdc_read_operation : Cdc_write_operation) |
      Cdc_stream_block | Cdc_init_sequence);

  // Pre-JZ4780 SoCs enable DMA in the command/data register.

  if (have_dma_enable_in_command())
    set_field(Msc_command_data_control, Cdc_dma_field_mask, Cdc_dma_field_shift,
              data ? Cdc_dma_enable : Cdc_dma_disable);

  // Unmask interrupts, start the clock, then initiate the command.

  uint32_t flags = Int_end_command_response | Int_response_timeout;

  unmask_irq(flags);
  start_clock();

  _regs[Msc_control] = _regs[Msc_control] | Control_start_operation;

  // Wait for command completion.

  if (!wait_for_irq(flags))
    return false;

  // Determine whether a timeout occurred.

  bool have_response = !((_regs[Msc_interrupt_flag] & Int_response_timeout) ||
                         (_regs[Msc_status] & Status_response_crc_error) ||
                         (_regs[Msc_status] & Status_timeout_response));

  // Acknowledge the response interrupts and return the status.

  ack_irq(flags);
  return have_response;
}

// Wait indefinitely for an interrupt request, returning true if one was delivered.

bool
Msc_channel::wait_for_irq(uint32_t flags)
{
  if (l4_error(l4_rcv_ep_bind_thread(_irq, get_current_thread(), 0)))
    return false;

  return !l4_error(l4_irq_receive(_irq, L4_IPC_NEVER)) &&
         (_regs[Msc_interrupt_flag] & flags);
}

// Wait up to the given timeout (in microseconds) for an interrupt request,
// returning true if one was delivered.

bool
Msc_channel::wait_for_irq(uint32_t flags, unsigned int timeout)
{
  if (l4_error(l4_rcv_ep_bind_thread(_irq, get_current_thread(), 0)))
    return false;

  return !l4_error(l4_irq_receive(_irq, l4_timeout(L4_IPC_TIMEOUT_NEVER,
                                                   l4util_micros2l4to(timeout)))) &&
         (_regs[Msc_interrupt_flag] & flags);
}



// Check the voltage range of the SD card, potentially establishing that it is
// a high capacity card. Return false if the voltage range is incompatible.

bool
Msc_channel::check_sd()
{
  struct R7 r7;

  // Send an interface condition command.
  // A card may not respond to this command.

  if (!send_command(Command_send_if_cond, If_cond_default_voltage_range))
    return true;

  read_response((uint16_t *) &r7, Response_size_R7);

  // Reject any card not supporting the default voltage range.

  if (r7.check_voltage.raw != If_cond_default_voltage_range)
    return false;

  return true;
}

// Check the voltage range of the SDIO card, inactivating it if incompatible.

void
Msc_channel::init_sdio()
{
  struct R4 r4;
  uint32_t ocr = 0;

  // Reset any SDIO card or IO unit in a combined memory/IO card.
  // A non-SDIO card may not respond to this command.

  if (!send_command(Command_io_rw_direct, Io_rw_direct_reset))
    return;

  // Attempt to assert the operating conditions.

  do
  {
    // Obtain OCR (operating conditions register) values for any IO card.
    // Without a response, the card may have inactivated itself due to voltage
    // range incompatibility reasons.

    if (!send_command(Command_io_send_op_cond, ocr))
      return;

    read_response((uint16_t *) &r4, Response_size_R4);

    // Finish if no IO functions provided.
    // NOTE: Should only need to check this the first time.

    if (r4.number_io_functions == 0)
      return;

    if (r4.ocr != Ocr_default_voltage_range)
    {
      ocr = Ocr_default_voltage_range;
      continue;
    }
  }
  while (!r4.ready);
}

void
Msc_channel::init_sdmem()
{
  struct R3 r3;

  // Incorporate the HCS bit into the OCR for SDMEM.

  uint32_t ocr = Ocr_high_capacity_storage;

  do
  {
    if (!send_app_command(App_command_sd_send_op_cond, ocr))
      return;

    read_response((uint16_t *) &r3, Response_size_R3);

    if (r3.ocr != Ocr_default_voltage_range)
    {
      ocr = Ocr_default_voltage_range | Ocr_high_capacity_storage;
      continue;
    }
  }
  while (!(r3.ocr & Ocr_card_powered_up));
}

void
Msc_channel::init_mmc()
{
  // Obtain OCR (operating conditions register) values for each card using
  // send_op_cond command variants without argument, or assert operating
  // conditions with argument to avoid handling card responses. Where responses
  // are solicited, the host must determine a suitable argument and reissue the
  // command.

  struct R3 r3;
  uint32_t ocr = 0;

  do
  {
    if (!send_command(Command_send_op_cond, ocr))
      return;

    read_response((uint16_t *) &r3, Response_size_R3);

    if (r3.ocr != Ocr_default_voltage_range)
    {
      ocr = Ocr_default_voltage_range;
      continue;
    }
  }
  while (!(r3.ocr & Ocr_card_powered_up));
}

void
Msc_channel::identify_cards()
{
  struct R2 r2;
  struct R6 r6;

  _num_cards = 0;

  while (send_command(Command_all_send_cid, 0))
  {
    read_response((uint16_t *) &r2, Response_size_R2);

    _cards[_num_cards].cid = r2.payload.cid;

    printf("card: %d\n", _num_cards);
    printf("date: %d %d\n", r2.payload.cid.month, r2.payload.cid.year);
    printf("serial: %d\n", r2.payload.cid.serial);
    printf("revision: %d\n", r2.payload.cid.revision);
    printf("name: %c%c%c%c%c\n", r2.payload.cid.name[4], r2.payload.cid.name[3],
                                 r2.payload.cid.name[2], r2.payload.cid.name[1],
                                 r2.payload.cid.name[0]);
    printf("oem: %d\n", r2.payload.cid.oem);
    printf("manufacturer: %d\n", r2.payload.cid.manufacturer);

    // Try and obtain a card-issued address.

    if (send_command(Command_send_relative_addr, 0))
    {
      read_response((uint16_t *) &r6, Response_size_R6);
      _cards[_num_cards].rca = r6.rca;
    }

    // Try and assign an address.
    // Employ 1-based relative addressing.

    else if (send_command(Command_set_relative_addr, _num_cards + 1))
      _cards[_num_cards].rca = _num_cards + 1;

    // Otherwise, stop identification.

    else
      return;

    // Set the default bus width to be determined.

    _cards[_num_cards].bus_width = 0;

    _num_cards++;
  }
}

void
Msc_channel::query_cards()
{
  struct R2 r2;
  struct R3 r3;
  uint8_t card;

  for (card = 0; card < _num_cards; card++)
  {
    // Employ 1-based relative addressing.

    if (!send_command(Command_send_csd, _cards[card].rca << 16))
      return;

    read_response((uint16_t *) &r2, Response_size_R2);

    _cards[card].csd = r2.payload.csd;

    struct CSD *csd = &_cards[card].csd;

    printf("card: %d\n", card);
    printf("csd: %d\n", csd->csd);
    printf("copy: %s\n", csd->copy ? "copied" : "original");
    printf("card command classes: %03x\n", csd->card_command_classes);
    printf("device (size multiplier): %d %d\n", csd->device_size + 1,
                                                1 << (csd->device_size_multiplier + 2));
    printf("device size: %d\n", (1 << csd->read_blocklen) *
                                (csd->device_size + 1) *
                                (1 << (csd->device_size_multiplier + 2)));
    printf("transfer speed: %d MHz\n", csd->tran_speed == 0x32 ? 25 : 50);
    printf("format group: %d %d\n", csd->format, r2.payload.csd.format_group);
    printf("write time factor: %d\n", 1 << csd->write_time_factor);
    printf("write protect (temp perm): %s %s\n", csd->temp_write_prot ? "yes" : "no",
                                                 csd->perm_write_prot ? "yes" : "no");
    printf("write protect group (enable size): %s %d\n", csd->write_prot_group_enable ? "yes" : "no",
                                                         csd->write_prot_group_size + 1);
    printf("write block (partial length): %s %d\n", csd->write_block_partial ? "yes" : "no",
                                                    1 << csd->write_blocklen);
    printf("read block (partial length): %s %d\n", csd->read_block_partial ? "yes" : "no",
                                                   1 << csd->read_blocklen);
    printf("erase: sector single: %d %s\n", csd->erase_sector_size + 1,
                                            csd->erase_single_block_enable ? "yes" : "no");
    printf("misalign: read write: %s %s\n", csd->read_block_misalign ? "yes" : "no",
                                           csd->write_block_misalign ? "yes" : "no");
    printf("max read current (min max): %d %d\n", csd->max_read_current_min,
                                                  csd->max_read_current_max);
    printf("max write current (min max): %d %d\n", csd->max_write_current_min,
                                                   csd->max_write_current_max);
    printf("read access time (1 2): %d %d\n", csd->data_read_access_time_1,
                                              csd->data_read_access_time_2);
    printf("DSR: %s\n", csd->dsr_implemented ? "yes" : "no");

    // Query the OCR again now that we can associate it with a specific card.

    if (!send_app_command(App_command_read_ocr, 0))
      return;

    read_response((uint16_t *) &r3, Response_size_R3);

    _cards[card].ocr = r3.ocr;
  }
}



// Enable the controller and identify cards.

void 
Msc_channel::enable()
{
  // NOTE: X1600 and other recent SoCs only.

  _regs[Msc_low_power_mode] = _regs[Msc_low_power_mode] & ~Low_power_mode_enable;

  stop_clock();
  reset();

  // Slow the clock for initialisation.
  // NOTE: Should produce an error.

  if (!set_clock_frequency(400000))
    return;

  send_command(Command_go_idle_state, 0);

  if (check_sd())
  {
    init_sdio();
    init_sdmem();
  }

  init_mmc();
  identify_cards();
  query_cards();

  // Initially, no card is selected.

  _card = -1;
}

// Obtain the card details.

struct msc_card *
Msc_channel::get_cards()
{
  return _cards;
}

// Return the number of active cards.

uint8_t
Msc_channel::num_cards()
{
  return _num_cards;
}

// Receive data from the selected card.

uint32_t
Msc_channel::recv_data(struct dma_region *region, uint32_t count)
{
  if (count > region->size)
    return 0;

  uint32_t flags = Int_data_transfer_done;

  unmask_irq(flags);

  uint32_t to_transfer = transfer(_msc_start + Msc_recv_data_fifo, region->paddr, true, count);

  wait_for_irq(flags);
  ack_irq(flags);

  if (!to_transfer ||
      (_regs[Msc_status] & Status_read_crc_error) ||
      (_regs[Msc_status] & Status_timeout_read))
    return 0;

  return to_transfer;
}

// Send data to the selected card.

uint32_t
Msc_channel::send_data(struct dma_region *region, uint32_t count)
{
  if (count > region->size)
    return 0;

  uint32_t flags = Int_data_transfer_done;

  unmask_irq(flags);

  uint32_t to_transfer = transfer(region->paddr, _msc_start + Msc_trans_data_fifo, false, count);

  wait_for_irq(flags);
  ack_irq(flags);

  if (!to_transfer ||
      (_regs[Msc_status] & Status_write_crc_error_data) ||
      (_regs[Msc_status] & Status_write_crc_error_no_status))
    return 0;

  return to_transfer;
}

// Read data from the indicated card into a memory region.

uint32_t 
Msc_channel::read(uint8_t card, struct dma_region *region,
                  uint32_t address, uint32_t count)
{
  uint32_t block_size = 1 << _cards[card].csd.read_blocklen;

  // Obtain the block address from the byte address and the number of blocks
  // from the byte count.

  // NOTE: Rejecting addresses and counts that are not multiples of the block
  // NOTE: size. Should produce an error.

  if ((address % block_size) || (count % block_size))
    return 0;

  return read_blocks(card, region, address / block_size, count / block_size);
}

// Read blocks from the indicated card into a memory region.

uint32_t 
Msc_channel::read_blocks(uint8_t card, struct dma_region *region,
                         uint32_t block_address, uint32_t block_count)
{
  uint32_t block_size = 1 << _cards[card].csd.read_blocklen;
  struct R1 r1;

  // Select the requested card.

  if (_card != card)
  {
    // Choose an appropriate clock frequency for the card.
    // NOTE: Should produce an error.

    if (!set_clock_frequency(_cards[card].csd.tran_speed == 0x32 ?
                             25000000 : 50000000))
      return 0;

    if (!send_command(Command_select_deselect_card, _cards[card].rca << 16))
      return 0;

    read_response((uint16_t *) &r1, Response_size_R1);

    if (r1.status & R1_status_error_mask)
      return 0;

    _card = card;
  }

  // NOTE: SMEM cards should allow bus width setting with the SCR register
  // NOTE: describing the permitted values.
  // NOTE: SDIO cards have their bus width set in CCCR via CMD52.

  if (!_cards[card].bus_width)
  {
    if (send_app_command(App_command_set_bus_width, Bus_width_4bit))
      _cards[card].bus_width = 4;
    else
      _cards[card].bus_width = 1;
  }

  if (!send_command(Command_set_blocklen, block_size))
    return 0;

  read_response((uint16_t *) &r1, Response_size_R1);

  if (r1.status & R1_status_error_mask)
    return 0;

  // NOTE: Consider issuing a predefined block count command to any multiple
  // NOTE: block operation, at least for cards that support it.

  // Apply block count and size properties to the issued command.

  _regs[Msc_block_count] = block_count;
  _regs[Msc_block_length] = block_size;

  // Where CCS = 0, byte addressing is used. Otherwise, block addressing is used.

  uint32_t address;

  if (_cards[card].ocr & Ocr_high_capacity_storage)
    address = block_address;
  else
    address = block_address * block_size;

  if (!send_command(block_count == 1 ? Command_read_single_block
                                     : Command_read_multiple_block, address))
    return 0;

  read_response((uint16_t *) &r1, Response_size_R1);

  if (r1.status & R1_status_error_mask)
    return 0;

  // NOTE: Use Msc_block_success_count instead.

  uint32_t transferred = recv_data(region, block_size * block_count);

  if (block_count > 1)
    send_command(Command_stop_transmission, 0);

  return transferred;
}



// Peripheral abstraction.

Msc_chip::Msc_chip(l4_addr_t msc_start, l4_addr_t start, l4_addr_t end,
                   Cpm_chip *cpm)
: _msc_start(msc_start), _start(start), _end(end), _cpm(cpm)
{
}
